## Code to apply any correction method:
## simple bias adjustment
## variance inflation
## quantile mapping
#
### Which parameter are required?
#if (!("obs" %in% ls()) || is.null(obs)) {
#  error(logger, 
#        "There is no object 'obs' in the global environment or it is NULL")
#  stop("EXECUTION FAILED")  
#}
#if (stream == "fcst" && (!("fcst" %in% ls()) || is.null(fcst))) {
#  error(logger, 
#        "There is no object 'fcst' in the global environment or it is NULL")
#  stop("EXECUTION FAILED")
#}
#if (!("hcst" %in% ls()) || is.null(hcst)) {
#  error(logger, 
#        "There is no object 'hcst' in the global environment or it is NULL")
#  stop("EXECUTION FAILED")
#}
#if (!("method" %in% ls()) || is.null(method)) {
#  warn(logger,
#       "Calibration method not found and it is set as 'SBC'.")
#  method <- 'SBC'
#}
#if (method %in% c('SBC', 'SimpleBiasCorrection')) {
#  cal_fun <- "CSTools::Calibration"
#  cal_method <- "bias"
#} else if (method %in% c("Inflation", "VarianceInflation")) {
#  cal_fun <- "CSTools::Calibration"
#  cal_method <- "evmos"
#} else if (method %in% c("QM", "QuantileMapping")) {
#  cal_fun <- "CSTools::QuantileMapping"
#} else {
#  error(logger, "Unknown calibration method definde in the recipe.")
#  stop("EXECUTION FAILED")
#}
#info(logger, paste("#-------------------------- ", "\n",
#                   "running Calibration module", "\n",
#                   "it can call", cal_fun ))

#' Function that bias-adjust the S2S hindcast
#
#' this function bias adjust the hindcast data using 
#' the leave-one-out method and the specified fun
#
#'@param obs A numeric array with named dimensions,
#'           representing the observational data used
#'           in the bias-adjustment method.
#'           'sday','syear','ensemble' are mandatory dims

#'@param hcst A numeric array with named dimensions,
#'           representing the system hindcast data used
#'           in the bias-adjustment method.
#'           'sday','syear','ensemble' are mandatory dims

#'@param fun bias adjustment function 

#'@param mm TRUE if the experiment data is conformed by
#'           multiple systems (dat > 1)

#'@param ncores An integer indicating the number of cores to 
#' use for parallel computation. The default value is NULL.

#'@param target_dims Dims needed to do the calibration 
#'@param output_dims Output dims from the calib fun 

#'@return A numeric array with the bias adjusted hindcast,
#'        

CST_CALIB_METHODS <- c("bias", "evmos", "mse_min", "crps_min", "rpc-based")

hcst_calib <- function(obs, hcst, method, mm=F, na.rm=T,
                       split_factor=1, ncores=32)
{

  # Replicates observations for bias adjusting each 
  # system of the multi-model separately
  ## TODO: Adapt code for multi-model case
  if(mm){
    obs.mm <- obs
    for(dat in 1:(dim(hcst)['dat'][[1]]-1)){
        obs.mm <- abind(obs2, obs,
                        along = which(names(dim(obs)) == 'dat'))
    }
    names(dim(obs.mm)) <- names(dim(obs)) 
    obs <- obs.mm
    remove(obs.mm)
  }
  
  if (method %in% CST_CALIB_METHODS) {
    # Hcst calibration 
    hcst <- CSTools::CST_Calibration(hcst, obs,
                                     cal.method = method,
                                     eval.method = "leave-one-out",
                                     multi.model = mm,
                                     na.fill = TRUE,
                                     na.rm = na.rm,
                                     apply_to = NULL,
                                     alpha = NULL,
                                     memb_dim = "ensemble",
                                     sdate_dim = "syear",
                                     ncores = ncores)

    hcst$data[!is.finite(hcst$data)] <- NA

  } else if (method %in% c("s2s4e")) {
    hcst <- Apply(data=list(obs = obs$data, hcst = hcst$data),
                            extra_info = list(na.rm = na.rm),
                            target_dims = c('sday', 'syear', 'ensemble'),
                            output_dims = c('sday', 'syear', 'ensemble'),
			    na.rm = na.rm,
			    ncores = ncores,
			    fun = .cal)[[1]]

    hcst[!is.finite(hcst)] <- NA
  } else {
    ## TODO: Implement other calibration methods, implement logger
    stop("Calibration method is not implemented in CSTools")
   # error(logger, 
   #       "Calibration method is not implemented in CSTools")
  }

  # hcst$data[!is.finite(hcst$data)] <- NA
  ## TODO: Is this line necessary?
  # remove(hcst)

  # Merges all the ensembles from the different systems into
  # one single ensemble
  if(mm){
    hcst <- MergeDims(hcst, 
                      merge_dims = c('dat', 'ensemble'), 
                      rename_dim = 'ensemble')
    hcst <- drop_na_dim(hcst, 'ensemble')
  }

# # removes dat and var dimensions if needed
#  try(hcst<-Subset(hcst,
#                  c('var'), list(1), drop='selected'))
#  try(hcst<-Subset(hcst,
#                   c('dat'), list(1), drop='selected'))

  return(hcst)

}

#' Function that bias-adjust the S2S forecast
#
#' this function bias adjust the forecast data using 
#'  hindcast and observational data 
#
#'@param obs A numeric array with named dimensions,
#'           representing the observational data used
#'           in the bias-adjustment method.
#'           'sday','syear','ensemble' are mandatory dims

#'@param hcst A numeric array with named dimensions,
#'           representing the system hindcast data used
#'           in the bias-adjustment method.
#'           'sday','syear','ensemble' are mandatory dims

#'@param fcst A numeric array with named dimensions,
#'           representing the system forecast to be
#'           bias-adjusted.  
#'           'sday','syear','ensemble' are mandatory dims

#'@param fun bias adjustment function 

#'@param mm TRUE if the experiment data is conformed by
#'           multiple systems (dat > 1)

#'@param ncores An integer indicating the number of cores to 
#' use for parallel computation. The default value is NULL.

#'@param target_dims Dims needed to do the calibration 
#'@param output_dims Output dims from the calib fun 

#'@return A numeric array with the bias adjusted forecast,
#'        

## TODO:
## needs and update: look at hcst version
fcst_calib <- function(obs, hcst, fcst, fun, mm=F,
                       na.rm=T, split_factor=1, ncores=32,
		       ## TODO: establish target dims and output dims
                       target_dims=c('sday', 'syear', 'ensemble'),
                       output_dims=c('sday', 'syear', 'ensemble'))
{

  # Replicates observations for bias adjusting each 
  # system of the multi-model separately
  ## TODO: Adapt multi-model case
  if(mm){
    obs.mm <- obs
    for(dat in 1:(dim(hcst)['dat'][[1]]-1)){
        obs.mm <- abind(obs2,obs,
                          along=which(names(dim(obs)) == 'dat'))
    }
    names(dim(obs.mm)) <- names(dim(obs)) 
    obs <- obs.mm
    remove(obs.mm)
  }
 
  # Fcst Calibration
  ## TODO: Return s2dv_cube
  calibrated_fcst <- Apply(data=list(obs = obs$data, hcst = hcst$data, fcst = fcst$data),
                           extra_info = list(na.rm = na.rm),
                           target_dims = target_dims,
                           output_dims = output_dims,
                           na.rm = na.rm,
                           ncores = ncores,
                           fun = .fcstcal)[[1]]
  
  ## TODO: Restore this line
  # calibrated_fcst$data[!is.finite(calibrated_fcst$data)] <- NA
  calibrated_fcst[!is.finite(calibrated_fcst)] <- NA
  
  # Merges all the ensembles from the different systems into
  # one single ensemble
  ## TODO: Adapt multi-model case
  if(mm){
    calibrated_fcst <- MergeDims(calibrated_fcst, 
                         merge_dims=c('dat', 'ensemble'), 
                         rename_dim = 'ensemble')
    calibrated_fcst <- drop_na_dim(calibrated_fcst, 'ensemble')
  }

#  # removes dat and var dimensions if needed
#  try(calibrated_fcst<-Subset(calibrated_fcst,
#                              c('var'),list(1),drop='selected'))
#  try(calibrated_fcst<-Subset(calibrated_fcst,
#                              c('dat'),list(1),drop='selected'))

  return(calibrated_fcst)


}
